#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

from fractions import Fraction
from math import radians
from typing import TYPE_CHECKING

from ephyra.core import DataHelper
from ephyra.data import FT_HORIZONTAL, FS_HORIZONTAL_PLUS, RT_AZIMUTHAL, FS_VERTICAL_MINUS, Parameters, State, RT_POLAR
from ephyra.formula import horizontal_4_to_3_fov_coefficient, horizontal_fov_to_80_coefficient

if TYPE_CHECKING:
    from typing import Type, TypeVar

    HT = TypeVar('HT', bound=DataHelper)

dh_app1 = DataHelper(Fraction(16, 10), radians(90), Fraction(4, 3), FT_HORIZONTAL, FS_HORIZONTAL_PLUS,
                     RT_AZIMUTHAL, ltr_coefficient_transformation_formula=horizontal_4_to_3_fov_coefficient,
                     radians_per_count=radians(0.022))


def dh_sys_app1_factory(dh: Type[HT]) -> HT:
    return dh(Fraction(16, 10), radians(90), Fraction(4, 3), FT_HORIZONTAL, FS_HORIZONTAL_PLUS, RT_AZIMUTHAL,
              ltr_coefficient_transformation_formula=horizontal_4_to_3_fov_coefficient,
              radians_per_count=radians(0.022), counts_per_unit=800, screen_diagonal=24)


dh_sys_app1 = dh_sys_app1_factory(DataHelper)


def dh_sys_app1_p_factory(dh: Type[HT]) -> HT:
    return dh(Fraction(16, 10), radians(90), Fraction(4, 3), FT_HORIZONTAL, FS_HORIZONTAL_PLUS, RT_POLAR,
              ltr_coefficient_transformation_formula=horizontal_4_to_3_fov_coefficient,
              radians_per_count=radians(0.022), counts_per_unit=800, screen_diagonal=24)


dh_sys_app1_p = dh_sys_app1_p_factory(DataHelper)

dh_app2 = DataHelper(Fraction(16, 10), radians(100), Fraction(16, 10), FT_HORIZONTAL, FS_VERTICAL_MINUS,
                     RT_AZIMUTHAL, ltr_coefficient_transformation_formula=horizontal_fov_to_80_coefficient,
                     radians_per_count=radians(2.2278481012658227))

dh_sys_app2 = DataHelper(Fraction(16, 10), radians(100), Fraction(16, 10), FT_HORIZONTAL, FS_VERTICAL_MINUS,
                         RT_AZIMUTHAL, ltr_coefficient_transformation_formula=horizontal_fov_to_80_coefficient,
                         radians_per_count=radians(2.2278481012658227), counts_per_unit=800, screen_diagonal=24)

# TODO: investigate Hypothesis library for exhaustive discrete strategies <AP>

state = State(1., 1.)

params_collection = []

for rpc in (False, True):
    for cpu in (False, True):
        for sl in (False, True):
            params_collection.append(Parameters(Fraction(1, 1), FS_HORIZONTAL_PLUS, radians_per_count=1.,
                                                consider_app_input_data=rpc, counts_per_unit=1.,
                                                consider_system_input_data=cpu, screen_length=1.,
                                                consider_physical_screen_data=sl))



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

from dataclasses import dataclass
from fractions import Fraction

from ._types import *

FS_HORIZONTAL_PLUS: FOVScaling = 0
FS_VERTICAL_MINUS: FOVScaling = 1

FT_HORIZONTAL: FOVType = 0
FT_VERTICAL: FOVType = 1

RT_AZIMUTHAL: RotationType = 0
RT_POLAR: RotationType = 1


@dataclass(frozen=True, eq=False)
class Parameters:
    screen_aspect_ratio: Fraction
    fov_scaling: FOVScaling
    radians_per_count: float = .0
    counts_per_unit: float = .0
    screen_length: float = .0
    consider_app_input_data: bool = False
    consider_system_input_data: bool = False
    consider_physical_screen_data: bool = False


@dataclass(frozen=True, eq=False)
class FOVData:
    fov_horizontal: float
    fov_vertical: float


@dataclass(frozen=True, eq=False)
class State:
    linear_to_rotary_measure: float
    fov: float



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from fractions import Fraction


class ParametersTestCaseMixin:
    Parameters = None
    FS_HORIZONTAL_PLUS = None
    FS_VERTICAL_MINUS = None

    def _test_considerations(self, p: Parameters, consider_aid: bool, consider_sid: bool, consider_psd: bool):
        self.assertEqual(p.consider_app_input_data, consider_aid)
        self.assertEqual(p.consider_system_input_data, consider_sid)
        self.assertEqual(p.consider_physical_screen_data, consider_psd)

    def _test_parameters(self, p: Parameters):
        self.assertEqual(p.screen_aspect_ratio, Fraction(1, 2))
        self.assertEqual(p.fov_scaling, self.FS_HORIZONTAL_PLUS)
        self.assertAlmostEqual(p.radians_per_count, 3.1, 5)
        self.assertAlmostEqual(p.counts_per_unit, 4.1, 5)
        self.assertAlmostEqual(p.screen_length, 5.1, 5)
        self._test_considerations(p, True, True, True)

    def test_parameters(self):
        p = self.Parameters(screen_aspect_ratio=Fraction(1, 2), fov_scaling=self.FS_HORIZONTAL_PLUS,
                            radians_per_count=3.1, counts_per_unit=4.1, screen_length=5.1, consider_app_input_data=True,
                            consider_system_input_data=True, consider_physical_screen_data=True)
        self._test_parameters(p)

    def test_parameters2(self):
        p = self.Parameters(screen_aspect_ratio=Fraction(1, 2), fov_scaling=self.FS_VERTICAL_MINUS)
        self.assertEqual(p.screen_aspect_ratio, Fraction(1, 2))
        self.assertEqual(p.fov_scaling, self.FS_VERTICAL_MINUS)
        self.assertAlmostEqual(p.radians_per_count, .0, 5)
        self.assertAlmostEqual(p.counts_per_unit, .0, 5)
        self.assertAlmostEqual(p.screen_length, .0, 5)
        self._test_considerations(p, False, False, False)

    def test_parameters3(self):
        p = self.Parameters(screen_aspect_ratio=Fraction(1, 2), fov_scaling=self.FS_VERTICAL_MINUS, radians_per_count=3,
                            consider_app_input_data=True)
        self.assertEqual(p.screen_aspect_ratio, Fraction(1, 2))
        self.assertEqual(p.fov_scaling, self.FS_VERTICAL_MINUS)
        self._test_considerations(p, True, False, False)

    def test_parameters4(self):
        p = self.Parameters(screen_aspect_ratio=Fraction(1, 2), fov_scaling=self.FS_VERTICAL_MINUS, counts_per_unit=4,
                            consider_system_input_data=True)
        self.assertEqual(p.screen_aspect_ratio, Fraction(1, 2))
        self.assertEqual(p.fov_scaling, self.FS_VERTICAL_MINUS)
        self._test_considerations(p, False, True, False)

    def test_parameters5(self):
        p = self.Parameters(screen_aspect_ratio=Fraction(1, 2), fov_scaling=self.FS_HORIZONTAL_PLUS, screen_length=4,
                            consider_physical_screen_data=True)
        self.assertEqual(p.screen_aspect_ratio, Fraction(1, 2))
        self.assertEqual(p.fov_scaling, self.FS_HORIZONTAL_PLUS)
        self._test_considerations(p, False, False, True)

    def test_parameters6(self):
        p = self.Parameters(Fraction(1, 2), self.FS_HORIZONTAL_PLUS, 3.1, 4.1, 5.1, True, True, True)
        self._test_parameters(p)


class FOVDataTestCaseMixin:
    FOVData = None

    def test_fov_data(self):
        fd = self.FOVData(fov_horizontal=1.1, fov_vertical=2.1)
        self.assertAlmostEqual(fd.fov_horizontal, 1.1)
        self.assertAlmostEqual(fd.fov_vertical, 2.1)

    def test_fov_data2(self):
        fd = self.FOVData(1.1, 2.1)
        self.assertAlmostEqual(fd.fov_horizontal, 1.1)
        self.assertAlmostEqual(fd.fov_vertical, 2.1)


class StateTestCaseMixin:
    State = None

    def test_state(self):
        s = self.State(linear_to_rotary_measure=1.1, fov=2.1)
        self.assertAlmostEqual(s.linear_to_rotary_measure, 1.1)
        self.assertAlmostEqual(s.fov, 2.1)

    def test_state2(self):
        s = self.State(1.1, 2.1)
        self.assertAlmostEqual(s.linear_to_rotary_measure, 1.1)
        self.assertAlmostEqual(s.fov, 2.1)


class ConstantTestCaseMixin:
    FS_HORIZONTAL_PLUS = None
    FS_VERTICAL_MINUS = None
    FT_HORIZONTAL = None
    FT_VERTICAL = None
    RT_AZIMUTHAL = None
    RT_POLAR = None

    def test_constants(self):
        self.assertEqual(self.FS_HORIZONTAL_PLUS, 0)
        self.assertEqual(self.FS_VERTICAL_MINUS, 1)
        self.assertEqual(self.FT_HORIZONTAL, 0)
        self.assertEqual(self.FT_VERTICAL, 1)
        self.assertEqual(self.RT_AZIMUTHAL, 0)
        self.assertEqual(self.RT_POLAR, 1)



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

from abc import ABC, abstractmethod
from fractions import Fraction
from math import pi
from typing import TYPE_CHECKING

from ..data import State, Parameters
from ..formula import calculate_fov, convert_fov_to_aspect_ratio, screen_ratio_ltr_measure, \
    screen_distance_ltr_measure, radians_per_unit_measure, radians_for_distance_from_center, \
    radians_for_ratio_from_center, calculate_screen_width_height, rotation_ltr_measure
from ..util import for_rotation_type, parameters_consistent

if TYPE_CHECKING:
    from typing import Tuple, Callable

    from ..data import RotationType, FOVScaling, FOVType, FOVData

    FOVTransformationFormula = Callable[[float, FOVData, Parameters], float]
    LTRCoefficientTransformationFormula = Callable[[float, FOVData, FOVData, Parameters], float]


class DataHelper:

    def _create_state(self, fov: float, fov_aspect_ratio: Fraction, fov_type: FOVType, ltr_coefficient: float) \
            -> Tuple[State, FOVData]:
        if self._fov_t_f is not None:
            fov = self._fov_t_f(fov, self._st_ex, self._p)
        fov_d = calculate_fov(fov, fov_aspect_ratio, fov_type)
        if fov_aspect_ratio != self._p.screen_aspect_ratio:
            fov_d = convert_fov_to_aspect_ratio(fov_d, self._p.fov_scaling, self._p.screen_aspect_ratio)
        if self._ltr_c_f is not None:
            ltr_coefficient = self._ltr_c_f(ltr_coefficient, fov_d, self._st_ex, self._p)
        return State(ltr_coefficient, for_rotation_type(self._rt, fov_d.fov_horizontal, fov_d.fov_vertical)), fov_d

    def __init__(self, screen_aspect_ratio: Fraction, fov: float, fov_aspect_ratio: Fraction, fov_type: FOVType,
                 fov_scaling: FOVScaling, rt: RotationType, *, ltr_coefficient: float = 1,
                 fov_transformation_formula: FOVTransformationFormula = None,
                 ltr_coefficient_transformation_formula: LTRCoefficientTransformationFormula = None,
                 radians_per_count: float = None, counts_per_unit: float = None, screen_diagonal: float = None):
        self._def_fov_ar = fov_aspect_ratio
        self._def_fov_t = fov_type
        opt = {}
        if radians_per_count is not None:
            opt['consider_app_input_data'] = True
            opt['radians_per_count'] = radians_per_count
        if counts_per_unit is not None:
            opt['consider_system_input_data'] = True
            opt['counts_per_unit'] = counts_per_unit
        if screen_diagonal is not None:
            sw, sh = calculate_screen_width_height(screen_aspect_ratio, screen_diagonal)
            opt['consider_physical_screen_data'] = True
            opt['screen_length'] = for_rotation_type(rt, sw, sh)
        self._p = Parameters(screen_aspect_ratio, fov_scaling, **opt)
        self._fov_t_f = fov_transformation_formula
        self._ltr_c_f = ltr_coefficient_transformation_formula
        self._rt = rt
        self._st_ex = None
        # dangerous, pay attention to the initialized fields
        self._st, self._st_ex = self._create_state(fov, fov_aspect_ratio, fov_type, ltr_coefficient)

    @property
    def primary_state(self) -> State:
        return self._st

    @property
    def parameters(self) -> Parameters:
        return self._p

    def get_state_for_fov(self, fov: float, fov_aspect_ratio: Fraction = None, fov_type: FOVType = None,
                          ltr_coefficient: float = 1):
        if fov_aspect_ratio is None:
            fov_aspect_ratio = self._def_fov_ar
        if fov_type is None:
            fov_type = self._def_fov_t
        state, _ = self._create_state(fov, fov_aspect_ratio, fov_type, ltr_coefficient)
        return state


class _AbstractCalculator(ABC):

    def __init__(self, s1: State, p1: Parameters, p2: Parameters = None, val1: float = None, val2: float = None):
        if p2 is None:
            p2 = p1
        else:
            if not parameters_consistent(p1, p2):
                raise ValueError('p1 and p2 are inconsistent')
        self._p2 = p2
        if val2 is None:
            val2 = val1
        self._rm = self.measure_formula(val1, s1, p1)
        self._val2 = val2

    @staticmethod
    @abstractmethod
    def measure_formula(param: float, s: State, p: Parameters):
        pass

    def coefficient_for(self, s2: State) -> float:
        return self.measure_formula(self._val2, s2, self._p2) / self._rm

    def sensitivity_for(self, s1_sens: float, s2: State) -> Tuple[float, float]:
        c = self.coefficient_for(s2)
        return c, s1_sens * c


class FullRotationCalculator(_AbstractCalculator):

    def __init__(self, s1: State, p1: Parameters, p2: Parameters = None):
        super().__init__(s1, p1, p2, None, None)

    @staticmethod
    def measure_formula(_: float, s: State, p: Parameters):
        return rotation_ltr_measure(s, p)


class ScreenRatioCalculator(_AbstractCalculator):

    def __init__(self, ratio1: float, s1: State, p1: Parameters, p2: Parameters = None, ratio2: float = None):
        super().__init__(s1, p1, p2, ratio1, ratio2)

    @staticmethod
    def measure_formula(param: float, s: State, p: Parameters):
        return screen_ratio_ltr_measure(param, s, p)


class ScreenDistanceCalculator(_AbstractCalculator):

    def __init__(self, distance1: float, s1: State, p1: Parameters, p2: Parameters = None, distance2: float = None):
        if not p1.consider_physical_screen_data:
            raise ValueError('p1 lacks physical screen data')
        super().__init__(s1, p1, p2, distance1, distance2)

    @staticmethod
    def measure_formula(param: float, s: State, p: Parameters):
        return screen_distance_ltr_measure(param, s, p)


class DetailsCalculator:

    def __init__(self, s: State, p: Parameters):
        if not (p.consider_app_input_data and p.consider_system_input_data and p.consider_physical_screen_data):
            raise ValueError('p lacks data')
        self._s = s
        self._p = p
        self._rpu = radians_per_unit_measure(s, p)

    def full_rotation_units(self, sens: float) -> float:
        return 2 * pi / (self._rpu * sens)

    def screen_ratio_units(self, ratio: float, sens: float) -> float:
        return radians_for_ratio_from_center(ratio, self._s) / (self._rpu * sens)

    def screen_distance_units(self, distance: float, sens: float) -> float:
        return radians_for_distance_from_center(distance, self._s, self._p) / (self._rpu * sens)



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from fractions import Fraction
from math import radians
from unittest import TestCase

from ephyra.data import FT_HORIZONTAL
from ephyra.discovery import find_settings_fov_aspect_ratio


class FOVSettingsDiscoveryTestCase(TestCase):

    def test_fov_aspect_ratio_discovery(self):
        ratio, diff = find_settings_fov_aspect_ratio(radians(70), radians(81), FT_HORIZONTAL, Fraction(16, 10))
        self.assertEqual(ratio, Fraction(4, 3))
        self.assertLessEqual(diff, 0.05)

    # TODO: test with real data for FT_VERTICAL <AP>



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from fractions import Fraction
from math import radians

from ephyra.data import FS_HORIZONTAL_PLUS, FS_VERTICAL_MINUS, FT_VERTICAL, RT_AZIMUTHAL, RT_POLAR
from ..helper import dh_sys_app1_factory, dh_sys_app1_p_factory, dh_app1, dh_sys_app1, dh_app2, dh_sys_app2, \
    params_collection, state


class DataHelperTestCaseMixin:
    data_helper = None

    def _test1(self, p, sp, s1, sl, fs, f1):
        # parameters
        self.assertEqual(p.screen_aspect_ratio, Fraction(16, 10))
        self.assertAlmostEqual(p.screen_length, sl, 5)
        self.assertTrue(p.consider_physical_screen_data)
        self.assertAlmostEqual(p.counts_per_unit, 800, 5)
        self.assertTrue(p.consider_system_input_data)
        self.assertAlmostEqual(p.radians_per_count, .0003839724354387525, 8)
        self.assertTrue(p.consider_app_input_data)
        self.assertEqual(p.fov_scaling, FS_HORIZONTAL_PLUS)
        # primary state
        self.assertAlmostEqual(sp.linear_to_rotary_measure, 1, 5)
        self.assertAlmostEqual(sp.fov, fs, 5)
        # state
        self.assertAlmostEqual(s1.linear_to_rotary_measure, .25, 5)
        self.assertAlmostEqual(s1.fov, f1, 5)

    def _test2(self, p, sp, s1, fs, f1):
        # parameters
        self.assertEqual(p.screen_aspect_ratio, Fraction(16, 10))
        self.assertAlmostEqual(p.screen_length, .0, 5)
        self.assertFalse(p.consider_physical_screen_data)
        self.assertAlmostEqual(p.counts_per_unit, .0, 5)
        self.assertFalse(p.consider_system_input_data)
        self.assertAlmostEqual(p.radians_per_count, .0, 5)
        self.assertFalse(p.consider_app_input_data)
        self.assertEqual(p.fov_scaling, FS_VERTICAL_MINUS)
        # primary state
        self.assertAlmostEqual(sp.linear_to_rotary_measure, 0.9, 5)
        self.assertAlmostEqual(sp.fov, fs, 5)
        # state
        self.assertAlmostEqual(s1.linear_to_rotary_measure, 1, 5)
        self.assertAlmostEqual(s1.fov, f1, 5)

    def test_azimuthal_rotation_data(self):
        dh_sys_app = dh_sys_app1_factory(self.data_helper)
        s1 = dh_sys_app.get_state_for_fov(radians(52.8605), fov_aspect_ratio=Fraction(16, 10),
                                          ltr_coefficient=0.5)
        self._test1(dh_sys_app.parameters, dh_sys_app.primary_state, s1, 20.35195929612211,
                    radians(100.3888578154696), radians(52.8605))

    def test_azimuthal_rotation_data2(self):
        dh = self.data_helper(Fraction(16, 10), radians(73.74), Fraction(16, 9), FT_VERTICAL, FS_VERTICAL_MINUS,
                              RT_AZIMUTHAL, ltr_coefficient=0.9, fov_transformation_formula=lambda f, df, pr: f * 1.2)
        s1 = dh.get_state_for_fov(radians(35), Fraction(4, 3), FT_VERTICAL)
        self._test2(dh.parameters, dh.primary_state, s1, radians(119.98340377498069), radians(54.20843241712849))

    def test_polar_rotation_data(self):
        dh_sys_app1_p = dh_sys_app1_p_factory(self.data_helper)
        s1 = dh_sys_app1_p.get_state_for_fov(radians(52.8605), fov_aspect_ratio=Fraction(16, 10),
                                             ltr_coefficient=0.5)
        self._test1(dh_sys_app1_p.parameters, dh_sys_app1_p.primary_state, s1, 12.71997456007632,
                    radians(73.73979529168803), radians(34.51628653564565))

    def test_polar_rotation_data2(self):
        dh = self.data_helper(Fraction(16, 10), radians(73.74), Fraction(16, 9), FT_VERTICAL, FS_VERTICAL_MINUS,
                              RT_POLAR, ltr_coefficient=0.9, fov_transformation_formula=lambda f, df, pr: f * 1.2)
        s1 = dh.get_state_for_fov(radians(35), Fraction(4, 3), FT_VERTICAL)
        self._test2(dh.parameters, dh.primary_state, s1, radians(94.51984219934904), radians(35.477565197940905))

    def test_invalid_rotation_type(self):
        with self.assertRaises(ValueError):
            self.data_helper(Fraction(16, 10), radians(73.74), Fraction(16, 9), FT_VERTICAL, FS_VERTICAL_MINUS,
                             -5, ltr_coefficient=0.9, fov_transformation_formula=lambda f, df, pr: f * 1.2)


class _ParametersConsistencyMixin:
    calc_factory = None

    def test_data_consistency(self):
        for p1 in params_collection:
            for p2 in params_collection:
                with self.subTest(p1=p1, p2=p2):
                    if not hasattr(self, 'data_consistency_assumption') or self.data_consistency_assumption(p1):
                        if p1.consider_physical_screen_data == p2.consider_physical_screen_data \
                                and p1.consider_system_input_data == p2.consider_system_input_data \
                                and p1.consider_app_input_data == p2.consider_app_input_data:
                            self.calc_factory(p1, p2)
                        else:
                            with self.assertRaises(ValueError):
                                self.calc_factory(p1, p2)


class _ParametersSufficiencyMixin:
    calc_factory = None

    def test_data_sufficiency(self):
        for p in params_collection:
            with self.subTest(p=p):
                if self.data_consistency_assumption(p):
                    self.calc_factory(p, p)
                else:
                    with self.assertRaises(ValueError):
                        self.calc_factory(p, p)


class FullRotationCalculatorTestCaseMixin(_ParametersConsistencyMixin):
    FullRotationCalculator = None

    def test_full_rotation_coefficient(self):
        s2 = dh_app1.get_state_for_fov(radians(45))
        c = self.FullRotationCalculator(dh_app1.primary_state, dh_app1.parameters)
        self.assertAlmostEqual(c.coefficient_for(s2), 2, 5)

    def test_full_rotation_sens(self):
        s2 = dh_app2.primary_state
        c = self.FullRotationCalculator(dh_app1.primary_state, dh_app1.parameters, dh_app2.parameters)
        cf, sens = c.sensitivity_for(2, s2)
        self.assertAlmostEqual(cf, .007899999999999999, 5)
        self.assertAlmostEqual(sens, .0158, 5)

    def calc_factory(self, p1, p2):
        return self.FullRotationCalculator(state, p1, p2)


class ScreenRatioCalculatorTestCaseMixin(_ParametersConsistencyMixin):
    ScreenRatioCalculator = None

    def test_screen_ratio_coefficient(self):
        s2 = dh_app1.get_state_for_fov(radians(45))
        c = self.ScreenRatioCalculator(.3755782578608322, dh_app1.primary_state, dh_app1.parameters)
        self.assertAlmostEqual(c.coefficient_for(s2), .8717319501838238, 5)

    def test_screen_ratio_coefficient2(self):
        s2 = dh_app2.primary_state
        c = self.ScreenRatioCalculator(.3755782578608322, dh_app1.primary_state, dh_app1.parameters, dh_app2.parameters,
                                       .1877891289304161)
        self.assertAlmostEqual(c.coefficient_for(s2), .004107746005894921, 5)

    def test_screen_ratio_sens(self):
        s2 = dh_app2.get_state_for_fov(radians(40))
        c = self.ScreenRatioCalculator(.3755782578608322, dh_app2.primary_state, dh_app2.parameters)
        cf, sens = c.sensitivity_for(.0158, s2)
        self.assertAlmostEqual(cf, .8070339035709131, 5)
        self.assertAlmostEqual(sens, .01275113567642043, 5)

    def calc_factory(self, p1, p2):
        return self.ScreenRatioCalculator(1, state, p1, p2, 2)


class ScreenDistanceCalculatorTestCaseMixin(_ParametersConsistencyMixin, _ParametersSufficiencyMixin):
    ScreenDistanceCalculator = None

    def test_screen_distance_coefficient(self):
        s2 = dh_sys_app1.get_state_for_fov(radians(45))
        c = self.ScreenDistanceCalculator(3.93700787, dh_sys_app1.primary_state, dh_sys_app1.parameters)
        self.assertAlmostEqual(c.coefficient_for(s2), .8741948552471518, 5)

    def test_screen_distance_coefficient2(self):
        s2 = dh_sys_app2.primary_state
        c = self.ScreenDistanceCalculator(3.93700787, dh_sys_app1.primary_state, dh_sys_app1.parameters,
                                          dh_sys_app2.parameters, 3.93700787 / 2)
        self.assertAlmostEqual(c.coefficient_for(s2), .004118163897214368, 5)

    def test_screen_distance_sens(self):
        s2 = dh_sys_app2.get_state_for_fov(radians(40))
        c = self.ScreenDistanceCalculator(3.93700787, dh_sys_app2.primary_state, dh_sys_app2.parameters)
        cf, sens = c.sensitivity_for(.0158, s2)
        self.assertAlmostEqual(cf, .8095353430632674, 5)
        self.assertAlmostEqual(sens, .012790658420399625, 5)

    def calc_factory(self, p1, p2):
        return self.ScreenDistanceCalculator(1, state, p1, p2, 2)

    @staticmethod
    def data_consistency_assumption(p):
        return p.consider_physical_screen_data


class _ParametersSufficiencyMixin:
    calc_factory = None

    def test_data_sufficiency(self):
        for p in params_collection:
            with self.subTest(p=p):
                if self.data_consistency_assumption(p):
                    self.calc_factory(p, p)
                else:
                    with self.assertRaises(ValueError):
                        self.calc_factory(p, p)


class DetailsCalculatorTestCaseMixin(_ParametersSufficiencyMixin):
    DetailsCalculator = None

    def test_full_rotation_units(self):
        d = self.DetailsCalculator(dh_sys_app1.primary_state, dh_sys_app1.parameters)
        self.assertAlmostEqual(d.full_rotation_units(2), 10.227272727272728, 5)

    def test_screen_ratio_units(self):
        d = self.DetailsCalculator(dh_sys_app1.primary_state, dh_sys_app1.parameters)
        self.assertAlmostEqual(d.screen_ratio_units(.3755782578608322, 2), .6892272609337514, 5)

    def test_screen_distance_units(self):
        d = self.DetailsCalculator(dh_sys_app1.primary_state, dh_sys_app1.parameters)
        self.assertAlmostEqual(d.screen_distance_units(10, 2), 1.411991816731347, 5)

    def calc_factory(self, p1, _):
        return self.DetailsCalculator(state, p1)

    @staticmethod
    def data_consistency_assumption(p):
        return p.consider_physical_screen_data and p.consider_system_input_data and p.consider_app_input_data



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ._native import _READY as _NATIVE

if _NATIVE:
    from ._native import *
else:
    from ._python import *



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from fractions import Fraction
from math import radians, degrees

from ephyra.data import FT_HORIZONTAL, FT_VERTICAL, FS_VERTICAL_MINUS, FS_HORIZONTAL_PLUS, FOVData, State, Parameters
from ..helper import dh_app1, dh_sys_app1


class ScreenCalculationsTestCaseMixin:
    calculate_screen_width_height = None

    def test_screen_width_height_calculations(self):
        w, h = self.calculate_screen_width_height(Fraction(16, 10), 24)
        self.assertAlmostEqual(w, 20.35195929612211, 5)
        self.assertAlmostEqual(h, 12.71997456007632, 5)


class FOVManipulationTestCaseMixin:
    calculate_fov = None
    convert_fov_to_aspect_ratio = None

    def test_fov_calculation(self):
        fov = self.calculate_fov(radians(100), Fraction(16, 10), FT_HORIZONTAL)
        self.assertAlmostEqual(degrees(fov.fov_horizontal), 100, 2)
        self.assertAlmostEqual(degrees(fov.fov_vertical), 73.36, 2)

    def test_fov_calculation2(self):
        fov = self.calculate_fov(radians(74), Fraction(4, 3), FT_VERTICAL)
        self.assertAlmostEqual(degrees(fov.fov_horizontal), 90.27, 2)
        self.assertAlmostEqual(degrees(fov.fov_vertical), 74, 2)

    def test_fov_calculation3(self):
        with self.assertRaises(ValueError):
            self.calculate_fov(1, Fraction(2, 3), -1)

    def test_fov_aspect_ratio_conversion(self):
        fov_p = self.calculate_fov(radians(70), Fraction(4, 3), FT_HORIZONTAL)
        fov_desired = self.calculate_fov(radians(80), Fraction(16, 10), FT_HORIZONTAL)
        fov_converted = self.convert_fov_to_aspect_ratio(fov_p, FS_HORIZONTAL_PLUS, Fraction(16, 10))
        self.assertAlmostEqual(fov_converted.fov_horizontal, fov_desired.fov_horizontal, 2)
        self.assertAlmostEqual(fov_converted.fov_vertical, fov_desired.fov_vertical, 2)

    def test_fov_aspect_ratio_conversion2(self):
        fov_p = self.calculate_fov(radians(70), Fraction(4, 3), FT_HORIZONTAL)
        fov_desired = self.calculate_fov(radians(70), Fraction(16, 10), FT_HORIZONTAL)
        fov_converted = self.convert_fov_to_aspect_ratio(fov_p, FS_VERTICAL_MINUS, Fraction(16, 10))
        self.assertAlmostEqual(fov_converted.fov_horizontal, fov_desired.fov_horizontal, 2)
        self.assertAlmostEqual(fov_converted.fov_vertical, fov_desired.fov_vertical, 2)

    def test_fov_aspect_ratio_conversion3(self):
        with self.assertRaises(ValueError):
            self.convert_fov_to_aspect_ratio(FOVData(1.2, 3.4), -1, Fraction(3, 4))


class CalculatorFormulaTestCaseMixin:
    radians_per_unit_measure = None
    rotation_ltr_measure = None
    radians_for_ratio_from_center = None
    radians_for_distance_from_center = None
    screen_ratio_ltr_measure = None
    screen_distance_ltr_measure = None

    def test_radians_per_unit_measure(self):
        p = Parameters(Fraction(1, 2), FS_HORIZONTAL_PLUS, 0.5, 4, consider_app_input_data=True,
                       consider_system_input_data=True)
        s = State(3, 1)
        self.assertAlmostEqual(self.radians_per_unit_measure(s, p), 6, 5)

    def test_rotation_ltr_measure(self):
        p = Parameters(Fraction(1, 2), FS_HORIZONTAL_PLUS, 0.5, 4, consider_app_input_data=True,
                       consider_system_input_data=True)
        s = State(3, 1)
        self.assertAlmostEqual(self.rotation_ltr_measure(s, p), 1 / 6, 5)

    def test_radians_for_ratio_from_center(self):
        self.assertAlmostEqual(self.radians_for_ratio_from_center(.3755782578608322, dh_app1.primary_state),
                               .4234308319224209, 5)

    def test_radians_for_distance_from_center(self):
        self.assertAlmostEqual(self.radians_for_distance_from_center(3.93700787, dh_sys_app1.primary_state,
                                                                     dh_sys_app1.parameters), .43465788352661283, 5)

    def test_screen_ratio_ltr_measure(self):
        self.assertAlmostEqual(self.screen_ratio_ltr_measure(.3755782578608322, dh_app1.primary_state,
                                                             dh_app1.parameters), 1102.7636174940023, 5)

    def test_screen_distance_ltr_measure(self):
        self.assertAlmostEqual(self.screen_distance_ltr_measure(3.93700787, dh_sys_app1.primary_state,
                                                                dh_sys_app1.parameters), 1.415003537395672, 5)


class LtRCoefficientTransformationFormulaTestCaseMixin:
    horizontal_4_to_3_fov_coefficient = None
    horizontal_fov_to_80_coefficient = None

    def test_horizontal_4_to_3_fov_coefficient(self):
        self.assertAlmostEqual(self.horizontal_4_to_3_fov_coefficient(1, FOVData(radians(2), radians(3)), None,
                                                                      dh_app1.parameters), 1, 5)

    def test_horizontal_4_to_3_fov_coefficient2(self):
        p = Parameters(Fraction(16, 10), FS_VERTICAL_MINUS)
        self.assertAlmostEqual(self.horizontal_4_to_3_fov_coefficient(1, FOVData(radians(100), radians(50)),
                                                                      FOVData(radians(50), radians(25)), p), 2, 5)

    def test_horizontal_fov_to_80_coefficient(self):
        self.assertAlmostEqual(self.horizontal_fov_to_80_coefficient(1, FOVData(radians(160), radians(-5)), None,
                                                                     dh_app1.parameters), 2, 5)



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

from typing import TYPE_CHECKING

from .data import RT_AZIMUTHAL, RT_POLAR, Parameters

if TYPE_CHECKING:
    from .data import RotationType


def for_rotation_type(rt: RotationType, for_azimuthal, for_polar):
    if rt == RT_AZIMUTHAL:
        return for_azimuthal
    elif rt == RT_POLAR:
        return for_polar
    else:
        raise ValueError(f'unknown rotation type {rt}')


def parameters_consistent(p1: Parameters, p2: Parameters):
    return p1.consider_app_input_data == p2.consider_app_input_data \
           and p1.consider_system_input_data == p2.consider_system_input_data \
           and p1.consider_physical_screen_data == p2.consider_physical_screen_data



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from setuptools import setup, Extension, find_packages

_ephyra = Extension('_ephyra', sources=['_ephyra/_ephyramodule.c', '_ephyra/data.c', '_ephyra/formula.c',
                                        '_ephyra/util.c', '_ephyra/core.c'])

setup(name='ephyra', version='0.1',
      ext_modules=[_ephyra], python_requires='>=3.8.1',
      packages=find_packages(exclude=['ephyra.tests', 'ephyra.tests.*']),
      description='Library to help with mouse input configuration',
      url='https://github.com/apleshakov/ephyra',
      author='Andrey Pleshakov',
      author_email='aplshkv@gmail.com',
      long_description_content_type='text/x-rst',
      license='Apache-2.0',
      classifiers=['Intended Audience :: Developers',
                   'License :: OSI Approved :: Apache Software License',
                   'Programming Language :: Python :: 3.8',
                   'Topic :: Software Development :: Libraries'])



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from unittest import TestCase

from ephyra.core._native import DataHelper, FullRotationCalculator, ScreenRatioCalculator, ScreenDistanceCalculator, \
    DetailsCalculator
from .common import DataHelperTestCaseMixin, FullRotationCalculatorTestCaseMixin, ScreenRatioCalculatorTestCaseMixin, \
    ScreenDistanceCalculatorTestCaseMixin, DetailsCalculatorTestCaseMixin


class NDataHelperTestCase(DataHelperTestCaseMixin, TestCase):
    data_helper = DataHelper


class NFullRotationCalculatorTestCase(FullRotationCalculatorTestCaseMixin, TestCase):
    FullRotationCalculator = FullRotationCalculator


class NScreenRatioCalculatorTestCase(ScreenRatioCalculatorTestCaseMixin, TestCase):
    ScreenRatioCalculator = ScreenRatioCalculator


class NScreenDistanceCalculatorTestCase(ScreenDistanceCalculatorTestCaseMixin, TestCase):
    ScreenDistanceCalculator = ScreenDistanceCalculator


class NDetailsCalculatorTestCase(DetailsCalculatorTestCaseMixin, TestCase):
    DetailsCalculator = DetailsCalculator



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from unittest import TestCase

from ephyra.data._python import Parameters, FS_HORIZONTAL_PLUS, FS_VERTICAL_MINUS, FOVData, State, FT_VERTICAL, \
    FT_HORIZONTAL, RT_AZIMUTHAL, RT_POLAR
from .common import ParametersTestCaseMixin, FOVDataTestCaseMixin, StateTestCaseMixin, ConstantTestCaseMixin


class PyParametersTestCase(ParametersTestCaseMixin, TestCase):
    Parameters = Parameters
    FS_HORIZONTAL_PLUS = FS_HORIZONTAL_PLUS
    FS_VERTICAL_MINUS = FS_VERTICAL_MINUS


class PyFOVDataTestCase(FOVDataTestCaseMixin, TestCase):
    FOVData = FOVData


class PyStateTestCase(StateTestCaseMixin, TestCase):
    State = State


class PyConstantsTestCase(ConstantTestCaseMixin, TestCase):
    FS_HORIZONTAL_PLUS = FS_HORIZONTAL_PLUS
    FS_VERTICAL_MINUS = FS_VERTICAL_MINUS
    FT_HORIZONTAL = FT_HORIZONTAL
    FT_VERTICAL = FT_VERTICAL
    RT_AZIMUTHAL = RT_AZIMUTHAL
    RT_POLAR = RT_POLAR



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

from fractions import Fraction
from math import tan, fabs, radians
from typing import TYPE_CHECKING

from .data import FT_HORIZONTAL, FT_VERTICAL

if TYPE_CHECKING:
    from typing import Tuple

    from .core import FOVType

_ASPECT_RATIOS = (
    Fraction(4, 3),
    Fraction(5, 4),
    Fraction(3, 2),
    Fraction(16, 10),
    Fraction(16, 9),
    Fraction(21, 9),
    Fraction(32, 9)
)


def find_settings_fov_aspect_ratio(fov_settings: float, fov_real: float, fov_type: FOVType, real_aspect_ratio: Fraction,
                                   ratios_to_test: Tuple[Fraction, ...] = _ASPECT_RATIOS) -> Tuple[Fraction, float]:
    if fov_type == FT_HORIZONTAL:
        r = tan(radians(fov_real) / 2) / tan(radians(fov_settings) / 2)
    elif fov_type == FT_VERTICAL:
        r = tan(radians(fov_settings) / 2) / tan(radians(fov_real) / 2)
    else:
        raise ValueError(f'unknown fov type {fov_type}')
    settings_ratio = float(real_aspect_ratio) / r
    return min(zip(ratios_to_test, (fabs(float(ar) - settings_ratio) for ar in ratios_to_test)), key=lambda t: t[1])



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

from fractions import Fraction
from math import atan, tan, sqrt, radians
from typing import TYPE_CHECKING

from ..data import Parameters, FT_HORIZONTAL, FT_VERTICAL, FS_HORIZONTAL_PLUS, FS_VERTICAL_MINUS, FOVData

if TYPE_CHECKING:
    from typing import Tuple

    from ..data import FOVScaling, FOVType, State


# core


def _fov_h(fov_v: float, aspect_ratio: Fraction) -> float:
    return 2 * atan(tan(fov_v / 2) * (aspect_ratio.numerator / aspect_ratio.denominator))


def _fov_v(fov_h: float, aspect_ratio: Fraction) -> float:
    return 2 * atan(tan(fov_h / 2) * (aspect_ratio.denominator / aspect_ratio.numerator))


def calculate_screen_width_height(aspect_ratio: Fraction, diagonal: float) -> Tuple[float, float]:
    unit_length = sqrt(diagonal ** 2 / (aspect_ratio.numerator ** 2 + aspect_ratio.denominator ** 2))
    return aspect_ratio.numerator * unit_length, aspect_ratio.denominator * unit_length


# TODO: Type/Value error consistency with native implementation <AP>

def calculate_fov(fov: float, aspect_ratio: Fraction, fov_type: FOVType) -> FOVData:
    if fov_type == FT_HORIZONTAL:
        specified_horizontal_fov = fov
        specified_vertical_fov = _fov_v(specified_horizontal_fov, aspect_ratio)
    elif fov_type == FT_VERTICAL:
        specified_vertical_fov = fov
        specified_horizontal_fov = _fov_h(specified_vertical_fov, aspect_ratio)
    else:
        raise ValueError(f'unknown fov type {fov_type}')
    return FOVData(specified_horizontal_fov, specified_vertical_fov)


def convert_fov_to_aspect_ratio(fov: FOVData, fov_scaling: FOVScaling,
                                aspect_ratio: Fraction) -> FOVData:
    if fov_scaling == FS_HORIZONTAL_PLUS:
        actual_vertical_fov = fov.fov_vertical
        actual_horizontal_fov = _fov_h(fov.fov_vertical, aspect_ratio)
    elif fov_scaling == FS_VERTICAL_MINUS:
        actual_horizontal_fov = fov.fov_horizontal
        actual_vertical_fov = _fov_v(fov.fov_horizontal, aspect_ratio)
    else:
        raise ValueError(f'unknown fov scaling {fov_scaling}')
    return FOVData(actual_horizontal_fov, actual_vertical_fov)


def radians_per_unit_measure(s: State, p: Parameters) -> float:
    v = s.linear_to_rotary_measure
    if p.consider_app_input_data:
        v *= p.radians_per_count
    if p.consider_system_input_data:
        v *= p.counts_per_unit
    return v


def rotation_ltr_measure(s: State, p: Parameters) -> float:
    return 1 / radians_per_unit_measure(s, p)


def radians_for_ratio_from_center(ratio: float, s: State) -> float:
    return atan(ratio * tan(s.fov / 2))


def screen_ratio_ltr_measure(ratio: float, s: State, p: Parameters) -> float:
    return radians_for_ratio_from_center(ratio, s) / radians_per_unit_measure(s, p)


def radians_for_distance_from_center(distance: float, s: State, p: Parameters) -> float:
    return atan(2 * distance * tan(s.fov / 2) / p.screen_length)


def screen_distance_ltr_measure(distance: float, s: State, p: Parameters):
    return radians_for_distance_from_center(distance, s, p) / radians_per_unit_measure(s, p)


# LtR Coefficients

_RAD_80 = radians(80)


def horizontal_4_to_3_fov_coefficient(b: float, c: FOVData, z: FOVData, p: Parameters) -> float:
    if z:
        c_fov = convert_fov_to_aspect_ratio(c, p.fov_scaling, Fraction(4, 3))
        z_fov = convert_fov_to_aspect_ratio(z, p.fov_scaling, Fraction(4, 3))
        return b * c_fov.fov_horizontal / z_fov.fov_horizontal
    else:
        return b


def horizontal_fov_to_80_coefficient(b: float, c: FOVData, _: FOVData, _2: Parameters) -> float:
    return b * c.fov_horizontal / _RAD_80
